# -*- coding: UTF-8 -*-
"""
Tests for the decorators module
"""

from __future__ import absolute_import, unicode_literals

try:
    from unittest.mock import Mock, call
except ImportError:
    from mock import Mock, call

from logging import getLogger
from inspect import isfunction
from sys import version_info
from time import sleep

import pytest

from pydecor.caches import FIFOCache, LRUCache, TimedCache
from pydecor.constants import LOG_CALL_FMT_STR
from pydecor.decorators import (
    after,
    before,
    log_call,
    construct_decorator,
    decorate,
    instead,
    intercept,
    memoize,
)


PY2 = version_info < (3, 0)


# Common parameters for testing decorators
# ret, kwargs, fn_args, fn_kwargs
# - ret is the return value of the function passed to the decorator
# - kwargs are the kwargs passed to the decoration call
# - fn_args are the args passed to the function
# - fn_kwargs are the kwargs passed to the function
#
# Each decorator type has a helper method, below, which determines what
# the expected call to the passed function will look like.
# Decorator-specific call logic is then tested in individual tests.
#
_base_params = (
    (None, {}, (), {}),
    ((('a', ), {'b': 'c'}), {}, ('d', ), {'e': 'f'}),
    (None, {'pow': 'bang'}, (), {}),
    ((('a', ), {'b': 'c'}), {'pow': 'bang'}, (), {}),
    (None, {'pow': 'bang', 'bop': 'smack'}, (), {}),
    (None, {'pass_decorated': True}, (), {}),
    ((('a', ), {'b': 'c'}), {'pass_decorated': True}, (), {}),
    (None, {'pass_params': False}, (), {}),
    ((('a', ), {'b': 'c'}), {'pass_params': False}, (), {}),
    (None, {'pass_params': False, 'pass_decorated': True}, (), {}),
    (None, {'pass_params': False}, ('a', ), {'b': 'c'}),
    (None, {'unpack_extras': False}, (), {}),
    ((('a', ), {'b': 'c'}), {'unpack_extras': False}, (), {}),
    (None, {'unpack_extras': False, 'pow': 'bang'}, (), {}),
    (None, {'unpack_extras': False, 'extras_key': 'boo', 'k': 'd'}, (), {}),
    (None, {}, (), {}),
    ((('a', ), {'b': 'c'}), {}, ('a', ), {'b': 'c'}),
    (None, {'pass_params': False}, (), {}),
    (None, {'pass_params': False, 'pass_decorated': False}, (), {}),
    (None, {'wowee': 'zappo'}, (), {}),
    (None, {'wowee': 'zappo', 'extras_key': 'zarblock'}, (), {}),
    (
        None,
        {'wowee': 'zappo', 'extras_key': 'zarblock', 'unpack_extras': False},
        (), {}
    ),
)


base_params = []
for i in _base_params:
    base_params.append((before, ) + i)
    base_params.append((after, ) + i)
    base_params.append((instead, ) + i)


# Extra parameter combinations mostly useful when testing class decoration
_class_params = (
    (None, {'implicit_method_decoration': False}, (), {}),
    (None, {'instance_methods_only': True}, (), {}),
    (
        None, 
        {'implicit_method_decoration': False, 'instance_methods_only': True}, 
        (), {}
    ),
    (
        None,
        {'implicit_method_decoration': True, 'instance_methods_only': False},
        (), {}
    ),
    ((('a', ), {'b': 'c'}), {'implicit_method_decoration': False}, (), {}),
    ((('a', ), {'b': 'c'}), {'instance_methods_only': True}, (), {}),
    (
        (('a', ), {'b': 'c'}),
        {'implicit_method_decoration': False, 'instance_methods_only': True},
        (), {}
    ),
    (
        (('a', ), {'b': 'c'}),
        {'implicit_method_decoration': True, 'instance_methods_only': False},
        (), {}
    ),
)

class_params = []
for i in _class_params:
    class_params.append((before, ) + i)
    class_params.append((after, ) + i)
    class_params.append((instead, ) + i)

class_params.extend(base_params)


def func_sig_before(kwargs, fn_args, fn_kwargs, decorated):
    """Return the expected signature for the decorator func"""
    exp_args = ()

    if kwargs.get('pass_params', False):
        exp_args += (fn_args, fn_kwargs)

    if kwargs.get('pass_decorated', False):
        exp_args += (decorated,)

    extras = {}

    for k, v in kwargs.items():
        if k not in instead.__code__.co_varnames:
            extras[k] = v

    exp_kwargs = {}

    if kwargs.get('unpack_extras', True):
        exp_kwargs = extras
    else:
        exp_kwargs[kwargs.get('extras_key', 'extras')] = extras

    return exp_args, exp_kwargs


def func_sig_after(kwargs, fn_args, fn_kwargs, decorated,
                   decorated_ret):
    """Return the expected signature for the decorator func"""
    exp_args = ()

    if kwargs.get('pass_result', True):
        exp_args += (decorated_ret, )

    if kwargs.get('pass_params', False):
        exp_args += (fn_args, fn_kwargs)

    if kwargs.get('pass_decorated', False):
        exp_args += (decorated, )

    extras = {}

    for k, v in kwargs.items():
        if k not in instead.__code__.co_varnames:
            extras[k] = v

    exp_kwargs = {}

    if kwargs.get('unpack_extras', True):
        exp_kwargs = extras
    else:
        exp_kwargs[kwargs.get('extras_key', 'extras')] = extras

    return exp_args, exp_kwargs


def func_sig_instead(kwargs, fn_args, fn_kwargs, decorated):
    """Return the expected signature for the decorator func"""
    exp_args = ()

    if kwargs.get('pass_params', True):
        exp_args += (fn_args, fn_kwargs)

    if kwargs.get('pass_decorated', True):
        exp_args += (decorated, )

    extras = {}

    for k, v in kwargs.items():
        if k not in instead.__code__.co_varnames:
            extras[k] = v

    exp_kwargs = {}

    if kwargs.get('unpack_extras', True):
        exp_kwargs = extras
    else:
        exp_kwargs[kwargs.get('extras_key', 'extras')] = extras

    return exp_args, exp_kwargs


func_sig_map = {
    before: func_sig_before,
    'before': func_sig_before,
    after: func_sig_after,
    'after': func_sig_after,
    instead: func_sig_instead,
    'instead': func_sig_instead,
}


@pytest.mark.parametrize('decor, ret, kwargs, fn_args, fn_kwargs', base_params)
def test_function_decoration(decor, ret, kwargs, fn_args, fn_kwargs):
    """Test the decoration of functions

    :param decor: the decorator to be used
    :param ret: the return value of the function to be passed to the
        decorator
    :param kwargs: the keyword arguments to be passed to the decorator
    :param fn_args: the arguments with which the decorated function
        will be called
    :param fn_kwargs: the kwargs with which the decorated function
        will be called
    """
    decorated_mock_return = 'billyrubin'

    func_mock = Mock(return_value=ret)
    decorated_mock = Mock(return_value=decorated_mock_return)

    func_mock.__name__ = str('func_mock')
    decorated_mock.__name__ = str('decorated_mock')

    fn = decor(func_mock, **kwargs)(decorated_mock)
    fn_ret = fn(*fn_args, **fn_kwargs)

    fn_sig_func = func_sig_map[decor]
    fn_sig_args = [kwargs, fn_args, fn_kwargs, decorated_mock]
    if decor is after:
        fn_sig_args.append(decorated_mock_return)

    exp_args, exp_kwargs = fn_sig_func(*fn_sig_args)

    func_mock.assert_called_once_with(*exp_args, **exp_kwargs)

    if decor is before:
        if ret is not None:
            assert len(ret) == 2
            args, kwargs = ret
            decorated_mock.assert_called_once_with(*args, **kwargs)
        else:
            decorated_mock.assert_called_once_with(*fn_args, **fn_kwargs)

        assert fn_ret == decorated_mock_return

    elif decor is after:
        decorated_mock.assert_called_once_with(*fn_args, **fn_kwargs)

        if ret is not None:
            assert fn_ret == ret
        else:
            assert fn_ret == decorated_mock_return

    elif decor is instead:
        decorated_mock.assert_not_called()
        assert ret == fn_ret

    else:
        assert False, 'Unexpected decorator???'


@pytest.mark.parametrize(
    'decor, ret, kwargs, fn_args, fn_kwargs', class_params
)
def test_class_decoration(decor, ret, kwargs, fn_args, fn_kwargs):
    """Test decorators when applied to a class"""
    func_mock = Mock(return_value=ret)
    decorated_return = 'wow, what a return'
    meth_mock = Mock(name='meth_mock', return_value=decorated_return)
    cls_mock = Mock(name='cls_mock', return_value=decorated_return)
    static_mock = Mock(name='static_mock', return_value=decorated_return)

    @decor(func_mock, **kwargs)
    class TheClass(object):

        def method(self, *args, **kwargs):
            return meth_mock(self, *args, **kwargs)

        @classmethod
        def cls(cls, *args, **kwargs):
            return cls_mock(cls, *args, **kwargs)

        @staticmethod
        def static(*args, **kwargs):
            return static_mock(*args, **kwargs)

    mocks = (meth_mock, cls_mock, static_mock)

    for name, mock in zip(('method', 'cls', 'static'), mocks):

        if (not kwargs.get('implicit_method_decoration', True) or
                (name != 'method' and
                     kwargs.get('instance_methods_only', False))):
            # This functionality tested elsewhere
            return

        theclass = TheClass()

        meth = getattr(theclass, name)
        meth_ret = meth(*fn_args, **fn_kwargs)

        exp_meth_args = ()

        if name == 'method':
            exp_meth_args += (theclass.__wrapped__, )
        elif name == 'cls':
            exp_meth_args += (TheClass.__wrapped__, )

        exp_meth_args += fn_args

        wrapped_meth = getattr(theclass.__wrapped__, name)

        fn_sig_func = func_sig_map[decor]
        fn_sig_args = [kwargs, fn_args, fn_kwargs, wrapped_meth]
        if decor is after:
            fn_sig_args.append(decorated_return)

        exp_args, exp_kwargs = fn_sig_func(*fn_sig_args)

        func_mock.assert_called_once_with(*exp_args, **exp_kwargs)
        func_mock.reset_mock()

        if decor is before:
            if ret is not None:
                assert len(ret) == 2
                exp_args, exp_kwargs = ret
                if name != 'static':
                    exp_args = (exp_meth_args[0], ) + exp_args
                mock.assert_called_once_with(*exp_args, **exp_kwargs)
            else:
                mock.assert_called_once_with(*exp_meth_args, **fn_kwargs)

            assert meth_ret == decorated_return

        elif decor is after:
            mock.assert_called_once_with(*exp_meth_args, **fn_kwargs)

            if ret is not None:
                assert meth_ret == ret
            else:
                assert meth_ret == decorated_return

        elif decor is instead:
            mock.assert_not_called()
            assert meth_ret == ret

        else:
            assert False, 'Unexpected decorator???'


@pytest.mark.parametrize('decor, ret, kwargs, fn_args, fn_kwargs', base_params)
def test_decorate_no_mixing(decor, ret, kwargs, fn_args, fn_kwargs,
                            pass_mock=None, decor_name=None):
    """Test the ``decorate`` decorator, one type at a time"""

    decorated_mock_return = 'Elementary'

    decor_name = decor_name or decor.__name__

    pass_mock = pass_mock or Mock(return_value=ret)

    decorated_mock = Mock(return_value=decorated_mock_return)

    pass_mock.__name__ = str('before_mock')
    decorated_mock.__name__ = str('decorated_mock')

    kwargs[decor_name] = pass_mock

    fn = decorate(**kwargs)(decorated_mock)

    fn_ret = fn(*fn_args, **fn_kwargs)

    fn_sig_func = func_sig_map[decor_name]
    fn_sig_args = [kwargs, fn_args, fn_kwargs, decorated_mock]
    if decor_name == 'after':
        fn_sig_args.append(decorated_mock_return)

    # Remove the before/after/instead key from the kwargs dict, since
    # it does not get passed from the decorator call down to the
    # before/after/instead call
    del kwargs[decor_name]

    exp_args, exp_kwargs = fn_sig_func(*fn_sig_args)

    pass_mock.assert_called_once_with(*exp_args, **exp_kwargs)

    if decor_name == 'before':
        if ret is not None:
            assert len(ret) == 2
            args, kwargs = ret
            decorated_mock.assert_called_once_with(*args, **kwargs)
        else:
            decorated_mock.assert_called_once_with(*fn_args, **fn_kwargs)

        assert fn_ret == decorated_mock_return

    elif decor_name == 'after':
        decorated_mock.assert_called_once_with(*fn_args, **fn_kwargs)

        if ret is not None:
            assert fn_ret == ret
        else:
            assert fn_ret == decorated_mock_return

    elif decor_name == 'instead':
        decorated_mock.assert_not_called()
        assert ret == fn_ret

    else:
        assert False, 'Unexpected decorator???'


@pytest.mark.parametrize('decor, ret, kwargs, fn_args, fn_kwargs', base_params)
def test_decorator_constructor_no_mixing(decor, ret, kwargs, fn_args,
                                         fn_kwargs):
    """Test that decorator-generated decorators behave as expected"""

    pass_mock = Mock(return_value=ret)

    pass_mock.__name__ = str('before_mock')

    kwargs[decor.__name__] = pass_mock

    # Make a custom generated decorator
    gen_decorator = construct_decorator(**kwargs)

    test_decorate_no_mixing(
        gen_decorator,
        ret,
        kwargs,
        fn_args,
        fn_kwargs,
        pass_mock=pass_mock,
        decor_name=decor.__name__,
    )


@pytest.mark.parametrize('decorator', [before, after, instead])
def test_direct_method_decoration(decorator):
    """Test the persistence across calls of extras"""

    inst_tracker = Mock(name='inst_tracker', return_value=None)
    cls_tracker = Mock(name='cls_tracker', return_value=None)
    static_tracker = Mock(name='static_tracker', return_value=None)

    inst_tracker.__name__ = str('inst_tracker')
    cls_tracker.__name__ = str('cls_tracker')
    static_tracker.__name__ = str('static_tracker')

    class SomeClass(object):

        @decorator(inst_tracker, pass_params=True)
        def some_method(self):
            pass

        @classmethod
        @decorator(cls_tracker, pass_params=True)
        def some_clsmethod(cls):
            pass

        @staticmethod
        @decorator(static_tracker, pass_params=True)
        def some_staticmethod():
            pass

    sc = SomeClass()

    sc.some_method(), sc.some_clsmethod(), sc.some_staticmethod()

    for mock in inst_tracker, cls_tracker, static_tracker:

        assert len(mock.mock_calls) == 1

        called_with = mock.call_args[0]

        if decorator is after:
            args, kwargs = called_with[1:3]
        else:
            args, kwargs = called_with[:2]

        assert args == ()
        assert kwargs == {}


@pytest.mark.parametrize('decorator', [before, after, instead])
def test_extras_persistence(decorator):
    """Test the persistence across calls of extras"""

    def memo_func(memo, **kwargs):
        memo.append('called')

    memo = []

    decorated = Mock(return_value=None)

    decorated.__name__ = str('decorated_mock')

    decorated = decorator(
        memo_func, pass_params=False, pass_result=False,
        pass_decorated=False, memo=memo,
    )(decorated)

    for _ in range(5):
        decorated()

    assert len(memo) == 5


@pytest.mark.parametrize('decorator', [before, after, instead])
def test_extras_persistence_class(decorator):
    """Test persistence of extras when decorating a class"""

    def memo_func(memo, **kwargs):
        memo.append('called')

    memo = []

    @decorator(memo_func, pass_params=False, pass_result=False,
               pass_decorated=False, memo=memo)
    class GreatClass(object):

        def awesome_method(self):
            pass

        @classmethod
        def classy_method(cls):
            pass

        @staticmethod
        def stately_method():
            pass

        @property
        def prop(self):
            return 'prop'

    gc = GreatClass()

    for _ in range(2):
        gc.awesome_method()

    assert len(memo) == 2

    assert gc.prop

    for _ in range(2):
        GreatClass.classy_method()

    assert len(memo) == 4

    for _ in range(2):
        gc.classy_method()

    assert len(memo) == 6

    for _ in range(2):
        GreatClass.stately_method()

    assert len(memo) == 8

    for _ in range(2):
        gc.stately_method()

    assert len(memo) == 10


@pytest.mark.parametrize('decorator', [before, after, instead])
def test_extras_persistence_class_inst_only(decorator):
    """Test persistence of extras, instance methods only"""

    def memo_func(memo, **kwargs):
        memo.append('called')

    memo = []

    @decorator(memo_func, pass_params=False, pass_result=False,
               pass_decorated=False, instance_methods_only=True,
               memo=memo)
    class GreatClass(object):

        def awesome_method(self):
            pass

        @classmethod
        def classy_method(cls):
            pass

        @staticmethod
        def stately_method():
            pass

        @property
        def prop(self):
            return 'prop'

    gc = GreatClass()

    for _ in range(2):
        gc.awesome_method()

    assert len(memo) == 2

    for _ in range(2):
        GreatClass.classy_method()

    assert gc.prop

    assert len(memo) == 2

    for _ in range(2):
        gc.classy_method()

    assert len(memo) == 2

    for _ in range(2):
        GreatClass.stately_method()

    assert len(memo) == 2

    for _ in range(2):
        gc.stately_method()

    assert len(memo) == 2


def test_stacking_before_after():
    """Test the stacking of before followed by after"""
    tracker = Mock(name='tracker', return_value=None)

    decorated = Mock(name='decorated', return_value='decorated')

    tracker.__name__ = str('tracker')
    decorated.__name__ = str('decorated')

    fn = before(tracker)(decorated)
    fn = after(tracker)(fn)

    fn()

    decorated.assert_called_once_with()

    assert tracker.call_count == 2
    before_call_args, before_call_kwargs = tracker.call_args_list[0]
    after_call_args, after_call_kwargs = tracker.call_args_list[1]

    assert before_call_args == ()
    assert before_call_kwargs == {}

    assert after_call_args == ('decorated', )
    assert after_call_kwargs == {}


def test_stacking_after_before():
    """Test stacking of after prior to before"""
    tracker = Mock(name='tracker', return_value=None)

    decorated = Mock(name='decorated', return_value='decorated')

    tracker.__name__ = str('tracker')
    decorated.__name__ = str('decorated')

    fn = after(tracker)(decorated)
    fn = before(tracker)(fn)

    fn()

    decorated.assert_called_once_with()

    assert tracker.call_count == 2

    # The order in which before/after are specified shouldn't matter
    before_call_args, before_call_kwargs = tracker.call_args_list[0]
    after_call_args, after_call_kwargs = tracker.call_args_list[1]

    assert before_call_args == ()
    assert before_call_kwargs == {}

    assert after_call_args == ('decorated', )
    assert after_call_kwargs == {}


@pytest.mark.parametrize('decor', [before, after])
def test_stacking_instead(decor):
    """Test stacking when instead is specified last (which is WRONG)

    Note that putting instead late in the stack WILL override
    any previous decorators!
    """
    tracker = Mock(name='tracker', return_value=None)

    decorated = Mock(name='decorated', return_value='decorated')

    tracker.__name__ = str('tracker')
    decorated.__name__ = str('decorated')

    fn = decor(tracker)(decorated)
    fn = instead(tracker)(fn)

    fn()

    decorated.assert_not_called()
    assert tracker.call_count == 1
    assert tracker.call_args[0][:2] == ((), {})
    assert isfunction(tracker.call_args[0][2])


def test_stacking_instead_before():
    """Test stacking instead prior to before

    In this case, before should run properly.
    """
    tracker = Mock(name='tracker', return_value=None)

    decorated = Mock(name='decorated', return_value='decorated')

    tracker.__name__ = str('tracker')
    decorated.__name__ = str('decorated')

    fn = instead(tracker)(decorated)
    fn = before(tracker)(fn)

    fn()

    decorated.assert_not_called()

    assert tracker.call_count == 2

    before_call_args, before_call_kwargs = tracker.call_args_list[0]
    instead_call_args, instead_call_kwargs = tracker.call_args_list[1]

    assert before_call_args == ()
    assert before_call_kwargs == {}

    assert instead_call_args == ((), {}, decorated)
    assert instead_call_kwargs == {}


def test_stacking_instead_after():
    """Test stacking instead prior to after

    In this case, after should behave properly
    """
    tracker = Mock(name='tracker', return_value=None)

    decorated = Mock(name='decorated', return_value='decorated')

    tracker.__name__ = str('tracker')
    decorated.__name__ = str('decorated')

    fn = instead(tracker)(decorated)
    fn = after(tracker)(fn)

    fn()

    decorated.assert_not_called()

    assert tracker.call_count == 2

    instead_call_args, instead_call_kwargs = tracker.call_args_list[0]
    after_call_args, after_call_kwargs = tracker.call_args_list[1]

    assert instead_call_args == ((), {}, decorated)
    assert instead_call_kwargs == {}

    assert after_call_args == (None, )
    assert after_call_kwargs == {}


def test_stacking_instead_before_after():
    """Test stacking all three decorators, instead-before-after

    As long as instead is specified first, this should work fine
    """
    tracker = Mock(name='tracker', return_value=None)

    decorated = Mock(name='decorated', return_value='decorated')

    tracker.__name__ = str('tracker')
    decorated.__name__ = str('decorated')

    fn = instead(tracker)(decorated)
    fn = before(tracker)(fn)
    fn = after(tracker)(fn)

    fn()

    decorated.assert_not_called()  # because instead() doesn't call it

    assert tracker.call_count == 3

    before_call_args, before_call_kwargs = tracker.call_args_list[0]
    instead_call_args, instead_call_kwargs = tracker.call_args_list[1]
    after_call_args, after_call_kwargs = tracker.call_args_list[2]

    assert before_call_args == ()
    assert before_call_kwargs == {}

    assert instead_call_args == ((), {}, decorated)
    assert instead_call_kwargs == {}

    assert after_call_args == (None, )
    assert after_call_kwargs == {}


def test_stacking_instead_after_before():
    """Test stacking all three decorators, instead-after-before

    This should work exactly the same as instead-before-after
    """
    tracker = Mock(name='tracker', return_value=None)

    decorated = Mock(name='decorated', return_value='decorated')

    tracker.__name__ = str('tracker')
    decorated.__name__ = str('decorated')

    fn = instead(tracker)(decorated)
    fn = after(tracker)(fn)
    fn = before(tracker)(fn)

    fn()

    decorated.assert_not_called()  # because instead() doesn't call it

    assert tracker.call_count == 3

    before_call_args, before_call_kwargs = tracker.call_args_list[0]
    instead_call_args, instead_call_kwargs = tracker.call_args_list[1]
    after_call_args, after_call_kwargs = tracker.call_args_list[2]

    assert before_call_args == ()
    assert before_call_kwargs == {}

    assert instead_call_args == ((), {}, decorated)
    assert instead_call_kwargs == {}

    assert after_call_args == (None, )
    assert after_call_kwargs == {}


@pytest.mark.parametrize('raises, catch, reraise, include_handler', [
    (Exception, Exception, ValueError, False),
    (Exception, Exception, ValueError, True),
    (Exception, Exception, True, True),
    (Exception, Exception, True, False),
    (None, Exception, ValueError, False),
    (None, Exception, ValueError, True),
    (Exception, Exception, None, False),
    (Exception, Exception, None, True),
    (Exception, RuntimeError, ValueError, False),  # won't catch
    (Exception, RuntimeError, ValueError, True),  # won't catch
])
def test_intercept(raises, catch, reraise, include_handler):
    """Test the intercept decorator"""
    wrapped = Mock()

    wrapped.__name__ = str('wrapped')

    if raises is not None:
        wrapped.side_effect = raises

    handler = Mock(name='handler') if include_handler else None

    if handler is not None:
        handler.__name__ = str('handler')

    fn = intercept(catch=catch, reraise=reraise, handler=handler)(wrapped)

    will_catch = raises and issubclass(raises, catch)

    if reraise and will_catch:
        to_be_raised = raises if reraise is True else reraise
        with pytest.raises(to_be_raised):
            fn()
    elif raises and not will_catch:
        with pytest.raises(raises):
            fn()
    else:
        fn()

    if handler is not None and will_catch:
        called_with = handler.call_args[0][0]
        assert isinstance(called_with, raises)

    if handler is not None and not will_catch:
        handler.assert_not_called()

    wrapped.assert_called_once_with(*(), **{})


def test_log_call():
    """Test the log_call decorator"""
    exp_logger = getLogger(__name__)
    exp_logger.debug = Mock()

    @log_call(level='debug')
    def func(*args, **kwargs):
        return 'foo'

    call_args = ('a', )
    call_kwargs = {'b': 'c'}

    call_res = func(*call_args, **call_kwargs)

    exp_msg = LOG_CALL_FMT_STR.format(
        name='func',
        args=call_args,
        kwargs=call_kwargs,
        result=call_res
    )

    exp_logger.debug.assert_called_once_with(exp_msg)


class TestMemoization:
    """Tests for memoization"""

    # (args, kwargs)
    memoizable_calls = (
        (('a', 'b'), {'c': 'd'}),
        ((['a', 'b', 'c'],), {'c': 'd'}),
        ((lambda x: 'foo',), {'c': lambda y: 'bar'}),
        (({'a': 'a'},), {'c': 'd'}),
        ((type(str('A'), (object,), {})(),), {}),
        ((), {}),
        ((1, 2, 3), {}),
    )

    @pytest.mark.parametrize('args, kwargs', memoizable_calls)
    def test_memoize_basic(self, args, kwargs):
        """Test basic use of the memoize decorator"""
        tracker = Mock(return_value='foo')

        @memoize()
        def func(*args, **kwargs):
            return tracker(args, kwargs)

        assert func(*args, **kwargs) == 'foo'
        tracker.assert_called_once_with(args, kwargs)

        assert func(*args, **kwargs) == 'foo'
        assert len(tracker.mock_calls) == 1

    def test_memoize_lru(self):
        """Test removal of least-recently-used items"""
        call_list = tuple(range(5))  # 0-4
        tracker = Mock()

        @memoize(keep=5, cache_class=LRUCache)
        def func(val):
            tracker(val)
            return val

        for val in call_list:
            func(val)

        # LRU: 0 1 2 3 4
        assert len(tracker.mock_calls) == len(call_list)
        for val in call_list:
            assert call(val) in tracker.mock_calls

        # call with all the same args
        for val in call_list:
            func(val)

        # no new calls, lru order should be same
        # LRU: 0 1 2 3 4
        assert len(tracker.mock_calls) == len(call_list)
        for val in call_list:
            assert call(val) in tracker.mock_calls

        # add new value, popping least-recently-used (0)
        # LRU: 1 2 3 4 5
        func(5)
        assert len(tracker.mock_calls) == len(call_list) + 1
        assert tracker.mock_calls[-1] == call(5)  # most recent call

        # Re-call with 0, asserting that we call the func again,
        # and dropping 1
        # LRU: 2 3 4 5 0
        func(0)
        assert len(tracker.mock_calls) == len(call_list) + 2
        assert tracker.mock_calls[-1] == call(0)  # most recent call

        # Let's ensure that using something rearranges it
        func(2)
        # LRU: 3 4 5 0 2
        # no new calls
        assert len(tracker.mock_calls) == len(call_list) + 2
        assert tracker.mock_calls[-1] == call(0)  # most recent call

        # Let's put another new value into the cache
        func(6)
        # LRU: 4 5 0 2 6
        assert len(tracker.mock_calls) == len(call_list) + 3
        assert tracker.mock_calls[-1] == call(6)

        # Assert that 2 hasn't been dropped from the list, like it
        # would have been if we hadn't called it before 6
        func(2)
        # LRU: 4 5 0 6 2
        assert len(tracker.mock_calls) == len(call_list) + 3
        assert tracker.mock_calls[-1] == call(6)

    def test_memoize_fifo(self):
        """Test using the FIFO cache"""
        call_list = tuple(range(5))  # 0-4
        tracker = Mock()

        @memoize(keep=5, cache_class=FIFOCache)
        def func(val):
            tracker(val)
            return val

        for val in call_list:
            func(val)

        # Cache: 0 1 2 3 4
        assert len(tracker.mock_calls) == len(call_list)
        for val in call_list:
            assert call(val) in tracker.mock_calls

        # call with all the same args
        for val in call_list:
            func(val)

        # no new calls, cache still the same
        # Cache: 0 1 2 3 4
        assert len(tracker.mock_calls) == len(call_list)
        for val in call_list:
            assert call(val) in tracker.mock_calls

        # add new value, popping first in (0)
        # Cache: 1 2 3 4 5
        func(5)
        assert len(tracker.mock_calls) == len(call_list) + 1
        assert tracker.mock_calls[-1] == call(5)  # most recent call

        # Assert 5 doesn't yield a new call
        func(5)
        assert len(tracker.mock_calls) == len(call_list) + 1
        assert tracker.mock_calls[-1] == call(5)  # most recent call

        # Re-call with 0, asserting that we call the func again,
        # and dropping 1
        # Cache: 2 3 4 5 0
        func(0)
        assert len(tracker.mock_calls) == len(call_list) + 2
        assert tracker.mock_calls[-1] == call(0)  # most recent call

        # Assert neither 0 nor 5 yield new calls
        func(0)
        func(5)
        assert len(tracker.mock_calls) == len(call_list) + 2
        assert tracker.mock_calls[-1] == call(0)  # most recent call

    def test_memoization_timed(self):
        """Test timed memoization"""
        time = 0.005
        tracker = Mock()

        @memoize(keep=time, cache_class=TimedCache)
        def func(val):
            tracker(val)
            return val

        assert func(1) == 1
        assert tracker.mock_calls == [call(1)]
        assert func(1) == 1
        assert tracker.mock_calls == [call(1)]
        sleep(time)
        assert func(1) == 1
        assert tracker.mock_calls == [call(1), call(1)]



# -*- coding: UTF-8 -*-
"""
Tests for the functions module
"""

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from functools import partial
from logging import getLogger


import pytest


from pydecor.decorators import Decorated
from pydecor.constants import LOG_CALL_FMT_STR
from pydecor.functions import (
    intercept,
    log_call
)


@pytest.mark.parametrize('raises, catch, reraise, include_handler', [
    (Exception, Exception, ValueError, False),
    (Exception, Exception, ValueError, True),
    (None, Exception, ValueError, False),
    (None, Exception, ValueError, True),
    (Exception, Exception, None, False),
    (Exception, Exception, None, True),
    (Exception, RuntimeError, ValueError, False),  # won't catch
    (Exception, RuntimeError, ValueError, True),  # won't catch
])
def test_interceptor(raises, catch, reraise, include_handler):
    """Test the intercept function"""
    wrapped = Mock()
    wrapped.__name__ = 'intercept_mock'
    if raises is not None:
        wrapped.side_effect = raises

    handler = Mock() if include_handler else None

    decorated = Decorated(wrapped, (), {})

    fn = partial(intercept, decorated, catch=catch, reraise=reraise,
                 handler=handler)

    will_catch = raises and issubclass(raises, catch)

    if reraise and will_catch:
        with pytest.raises(reraise):
            fn()
    elif raises and not will_catch:
        with pytest.raises(raises):
            fn()
    else:
        fn()

    if handler is not None and will_catch:
        called_with = handler.call_args[0][0]
        assert isinstance(called_with, raises)

    if handler is not None and not will_catch:
        handler.assert_not_called()

    wrapped.assert_called_once_with(*(), **{})


def test_log_call():
    """Test automatic logging"""
    exp_logger = getLogger(__name__)
    exp_logger.debug = Mock()

    def func(*args, **kwargs):
        return 'foo'

    call_args = ('a', )
    call_kwargs = {'b': 'c'}
    decorated = Decorated(func, call_args, call_kwargs)
    call_res = decorated(*decorated.args, **decorated.kwargs)

    log_call(decorated, level='debug')

    exp_msg = LOG_CALL_FMT_STR.format(
        name='func',
        args=call_args,
        kwargs=call_kwargs,
        result=call_res
    )

    exp_logger.debug.assert_called_once_with(exp_msg)



# -*- coding: UTF-8 -*-
"""
Generic decorators help you to easily create your own decorators with simple
``@before``, ``@after``, and ``@instead`` decorators that call a function
as specified in the execution order of the decorated callable.

Additionally the ``@decorate`` decorator allows you to combine the above three
generics in one call, and the ``construct_decorator`` function returns a
ready-to-use decorator implementing any or all of the above three generics,
which you can then assign to a variable name and use as desired.
"""

from __future__ import absolute_import, unicode_literals


__all__ = (
    'after',
    'before',
    'construct_decorator',
    'decorate',
    'instead',
    'Decorated',
    'DecoratorType',
)


from inspect import isclass
from logging import getLogger
from functools import partial, wraps
from types import FunctionType, MethodType
from typing import Union

from ._utility import ClassWrapper, get_fn_args


log = getLogger(__name__)


DecoratorType = Union[FunctionType, MethodType, type]


class Decorated(object):
    """A representation of a decorated class

    This user-immutable object provides information about the decorated
    class, method, or function. The decorated callable can be called
    by directly calling the ``Decorated`` instance, or via the
    ``wrapped`` instance attribute.

    The example below illustrates direct instantiation of the
    ``Decorated`` class, but generally you will only deal with
    instances of this class when they are passed to the functions
    specified on generic decorators.

    .. code:: python

        from pydecor import Decorated

        def some_function(*args, **kwargs):
            return 'foo'

        decorated = Decorated(some_function, ('a', 'b'), {'c': 'c'})

        assert decorated.wrapped.__name__ == some_function.__name__
        assert decorated.args == ('a', 'b')
        assert decorated.kwargs == {'c': 'c'}
        assert decorated.result is None  # has not yet been called

        res = decorated(decorated.args, decorated.kwargs)

        assert 'foo' == res == decorated.result

    .. note::

        identity tests ``decorated.wrapped is some_decorated_function``
        will not work on the ``wrapped`` attribute of a ``Decorated``
        instance, because internally the wrapped callable is wrapped
        in a method that ensures that ``Decorated.result`` is set
        whenever the callable is called. It is wrapped using
        ``functools.wraps``, so attributes like ``__name__``,
        ``__doc__``, ``__module__``, etc. should still be the
        same as on an actual reference.

        If you need to access a real reference to the wrapped
        function for any reason, you can do so by accessing
        the ``__wrapped__`` property, on ``wrapped``, which is
        set by ``functools.wraps``, e.g.
        ``decorated.wrapped.__wrapped__``.

    :param wrapped: a reference to the wrapped callable. Calling the
        wrapped callable via this reference will set the ``result``
        attribute.
    :param args: a tuple of arguments with which the decorated function
        was called
    :param kwargs: a dict of arguments with which the decorated function
        was called
    :param result: either ``None`` if the wrapped callable has not yet
        been called or the result of that call
    """

    __slots__ = ('args', 'kwargs', 'wrapped', 'result')

    def __init__(self, wrapped, args, kwargs, result=None):
        """Instantiate a Decorated object

        :param callable wrapped: the callable object being wrapped
        :param tuple args: args with which the callable was called
        :param dict kwargs: keyword arguments with which the callable
            was called
        :param
        """
        sup = super(Decorated, self)
        sup.__setattr__('args', get_fn_args(wrapped, args))
        sup.__setattr__('kwargs', kwargs)
        sup.__setattr__('wrapped', self._sets_results(wrapped))
        sup.__setattr__('result', result)

    def __str__(self):
        """Return a nice string of self"""
        if hasattr(self.wrapped, '__name__'):
            name = self.wrapped.__name__
        else:
            name = str(self.wrapped)
        return '<Decorated {}({}, {})>'.format(name, self.args, self.kwargs)

    def __call__(self, *args, **kwargs):
        """Call the function the specified arguments.

        Also set ``self.result``
        """
        return self.wrapped(*args, **kwargs)

    def __setattr__(self, key, value):
        """Disallow attribute setting"""
        raise AttributeError(
            'Cannot set "{}" because {} is immutable'.format(key, self)
        )

    def _sets_results(self, wrapped):
        """Ensure that calling ``wrapped()`` sets the result attr

        :param callable wrapped: the wrapped function, class, or method
        """
        @wraps(wrapped)
        def wrapped_wrapped(*args, **kwargs):
            """Set self.result after calling wrapped"""
            res = wrapped(*args, **kwargs)
            super(Decorated, self).__setattr__('result', res)
            return res

        return wrapped_wrapped


def before(func, pass_params=False, pass_decorated=False,
           implicit_method_decoration=True, instance_methods_only=False,
           unpack_extras=True, extras_key='extras',
           _use_future_syntax=False, **extras):
    """Specify a callable to be run before the decorated resource

    A callable provided to this decorator will be called
    any time the decorated function is executed, immediately before
    its execution.

    The callable is expected to either return ``None`` or to return
    a tuple and dict with which to replace the arguments to the
    decorated function.

    By default, the provided callable is called with no arguments
    and only any keyword arguments provided as extras to the
    decoration call::

        (**extras)

    However, this is configurable. If all possible ``pass``
    parameters are True, the call signature would look like this::

        (decorated_args, decorated_kwargs, decorated, **extras)

    Items that are omitted will be removed from the call signature.
    For example, specifying ``pass_params=False`` and
    ``pass_decorated=True`` would adjust the call signature to
    look like::

        (decorated, **extras)

    Specifying ``pass_params=True`` will pass the arguments and
    keyword arguments of the decorated function (as a tuple and
    a dict, respectively) to the provided callable.

    Specifying ``pass_decorated=True`` will pass a reference
    to the decorated function to the provided callable.

    Any extra keyword arguments passed to the decoration call
    will by default be passed directly as keyword arguments
    to ``func``. If keyword names conflict with the passed
    function, ``unpack_extras=False`` will cause extra keyword
    arguments to be passed as a dictionary using the value
    of ``extras_key``, which defaults to ``"extras"``.

    :param Callable func:
        a callable to run before the decorated function. By default,
        it is called with no arguments, but the call signature may
        be changed depending on the value of ``pass_params`` and
        ``pass_decorated``

    :param bool pass_params:
        (default False) if True, the arguments and keyword arguments to
        the decorated function will be passed to ``func`` as a tuple
        and a dict, respectively

    :param bool pass_decorated:
        (default False) if True, a reference to the decorated function
        will be passed to the provided ``func``

    :param bool implicit_method_decoration:
        (default True) if True, decorating a class implies decorating
        all of its methods

    :param bool instance_methods_only:
        (default False) if True, decorating a class implies decorating
        only its instance methods (not ``classmethod``- or
        ``staticmethod``- decorated methods)

    :param bool unpack_extras:
        (default True) if True, any extra keyword arguments included
        in the decorator call will be passed as keyword arguments
        to ``func``. If ``False``, extra keyword arguments will be
        passed to func as a dict assigned to the keyword argument
        corresponding to ``key``

    :param str extras_key:
        (default ``"extras"``) the key to which to assign extra
        decorator keyword arguments when ``unpack_extras`` is ``False``

    :param bool _use_future_syntax:
        (default False) if True, use the new ``Decorated`` object,
        which will become the default in version 2.0.0. If this
        parameter is True, all other parameters will be ignored
        expect for ``instance_methods_only`` and
        ``implicit_method_decoration``, a ``Decorated`` object
        will be passed to the function as its first argument,
        and any provided keyword arguments will be unpacked.

    :param **dict extras:
        any extra keyword arguments supplied to the decoration call

    :return: the decorated function/method/class
    :rtype: Union[FunctionType,MethodType,type]
    """

    def decorator(decorated):
        """The function that returns a replacement for the original"""

        def wrapper(*args, **kwargs):
            """The function that replaces the decorated one"""

            if unpack_extras or _use_future_syntax:
                fkwargs = extras
            else:
                fkwargs = {extras_key: extras}

            fn = func
            decor = decorated

            if _use_future_syntax:
                decor = Decorated(decorated, args, kwargs)
                fn = partial(fn, decor)
            else:
                if pass_params:
                    fn_args = get_fn_args(decorated, args)
                    fn = partial(fn, fn_args, kwargs)

                if pass_decorated:
                    fn = partial(fn, decorated)

            fret = fn(**fkwargs)

            if fret is not None:
                args, kwargs = fret

            ret = decor(*args, **kwargs)
            return ret

        if implicit_method_decoration and isclass(decorated):
            return ClassWrapper.wrap(
                decorated,
                decorator,
                instance_methods_only=instance_methods_only
            )

        # Equivalent to @wraps(decorated) on `wrapper`
        return wraps(decorated)(wrapper)

    return decorator


def after(func, pass_params=False, pass_result=True, pass_decorated=False,
          implicit_method_decoration=True, instance_methods_only=False,
          unpack_extras=True, extras_key='extras',
          _use_future_syntax=False, **extras):
    """Specify a callable to be run after the decorated resource

    A callable provided to this decorator will be called
    any time the decorated function is executed, immediately after
    its execution.

    If the callable returns a value, that value will replace the
    return value of the decorated function.

    By default, the provided callable is called with the return from
    the decorated function as its first argument and only any
    keyword arguments provided as extras to the decoration call::

        (decorated_result, **extras)

    However, this is configurable. If all possible ``pass``
    parameters are True, the call signature would look like this::

        (decorated_result, decorated_args,
         decorated_kwargs, decorated, **extras)

    Items that are omitted will be removed from the call signature.
    For example, specifying ``pass_params=False`` and
    ``pass_decorated=True`` would adjust the call signature to
    look like::

        (decorated_result, decorated, **extras)

    Specifying ``pass_result=False`` will prevent the result
    from being passed to the callable.

    Specifying ``pass_params=True`` will pass the arguments and
    keyword arguments (as a tuple and a dict) to the decorated
    function as the second and third arguments, respectively.

    Specifying ``pass_decorated=True`` will pass a reference
    to the decorated function as the second argument, or the
    fourth argument if ``pass_params=True``

    Any extra keyword arguments passed to the decoration call
    will by default be passed directly as keyword arguments
    to ``func``. If keyword names conflict with the passed
    function, ``unpack_extras=False`` will cause extra keyword
    arguments to be passed as a dictionary using the value
    of ``extras_key``, which defaults to ``"extras"``.

    :param Callable func:
        a callable to run after the decorated function. By default,
        it is called with the result of the decorated function as
        its only argument, but the call signature may be changed
        depending on the value of ``pass_params`` and ``pass_decorated``

    :param bool pass_result:
        (default True) if True, the return from the decorated
        function will be passed to the provided callable

    :param bool pass_params:
        (default False) if True, the arguments and keyword arguments to
        the decorated function will be passed to ``func`` as a tuple
        and a dict, respectively

    :param bool pass_decorated:
        (default False) if True, a reference to the decorated function
        will be passed to the provided ``func``

    :param bool implicit_method_decoration:
        (default True) if True, decorating a class implies decorating
        all of its methods

    :param bool instance_methods_only:
        (default False) if True, decorating a class implies decorating
        only its instance methods (not ``classmethod``- or
        ``staticmethod``- decorated methods)

    :param bool unpack_extras:
        (default True) if True, any extra keyword arguments included
        in the decorator call will be passed as keyword arguments
        to ``func``. If ``False``, extra keyword arguments will be
        passed to func as a dict assigned to the keyword argument
        corresponding to ``key``

    :param str extras_key:
        (default ``"extras"``) the key to which to assign extra
        decorator keyword arguments when ``unpack_extras`` is ``False``

    :param bool _use_future_syntax:
        (default False) if True, use the new ``Decorated`` object,
        which will become the default in version 2.0.0. If this
        parameter is True, all other parameters will be ignored
        expect for ``instance_methods_only`` and
        ``implicit_method_decoration``, a ``Decorated`` object
        will be passed to the function as its first argument,
        and any provided keyword arguments will be unpacked.

    :param **dict extras:
        any extra keyword arguments supplied to the decoration call

    :return: the decorated function/method/class
    :rtype: Union[FunctionType,MethodType,type]
    """

    def decorator(decorated):
        """The function that returns a replacement for the original"""

        def wrapper(*args, **kwargs):
            """The function that replaces the decorated one"""

            decor = decorated

            if _use_future_syntax:
                decor = Decorated(decorated, args, kwargs)

            ret = decor(*args, **kwargs)

            fn = func

            if _use_future_syntax:
                fn = partial(fn, decor)
            else:
                if pass_result:
                    fn = partial(fn, ret)

                if pass_params:
                    fn_args = get_fn_args(decorated, args)
                    fn = partial(fn, fn_args, kwargs)

                if pass_decorated:
                    fn = partial(fn, decorated)

            if unpack_extras or _use_future_syntax:
                fkwargs = extras
            else:
                fkwargs = {extras_key: extras}

            fret = fn(**fkwargs)

            if fret is not None:
                return fret

            return ret

        if implicit_method_decoration and isclass(decorated):
            return ClassWrapper.wrap(
                decorated,
                decorator,
                instance_methods_only=instance_methods_only
            )

        # Equivalent to @wraps(decorated) on `wrapper`
        return wraps(decorated)(wrapper)

    return decorator


def instead(func, pass_params=True, pass_decorated=True,
            implicit_method_decoration=True, instance_methods_only=False,
            unpack_extras=True, extras_key='extras',
            _use_future_syntax=False, **extras):
    """Specify a callable to be run in hte place of the decorated resource

    A callable provided to this decorator will be called any time the
    decorated function is executed. The decorated function **will not**
    be called unless the provided callable calls it.

    Whatever the provided callable returns is the return value of the
    decorated function.

    By default, the provided callable is called with the arguments
    and keyword arguments used to call the decorated function, as a
    tuple and a dict, respectively, and a reference to the decorated
    function. Any extra keyword arguments specified in the decoration
    call are also passed to the callable::

        (decorated_args, decorated_kwargs, decorated, **extras)

    Items that are omitted will be removed from the call signature.
    For example, specifying ``pass_params=False`` would adjust the
    call signature to look like::

        (decorated, **extras)

    Any extra keyword arguments passed to the decoration call
    will by default be passed directly as keyword arguments
    to ``func``. If keyword names conflict with the passed
    function, ``unpack_extras=False`` will cause extra keyword
    arguments to be passed as a dictionary using the value
    of ``extras_key``, which defaults to ``"extras"``.

    :param Callable func:
        a callable to run in place of the decorated function. By
        default, it is called the decorated function's arguments
        as a tuple, keyword arguments as a dict, and a reference
        to the decorated function

    :param bool pass_params:
        (default False) if True, the arguments and keyword arguments to
        the decorated function will be passed to ``func`` as a tuple
        and a dict, respectively

    :param bool pass_decorated:
        (default False) if True, a reference to the decorated function
        will be passed to the provided ``func``

    :param bool implicit_method_decoration:
        (default True) if True, decorating a class implies decorating
        all of its methods

    :param bool instance_methods_only:
        (default False) if True, decorating a class implies decorating
        only its instance methods (not ``classmethod``- or
        ``staticmethod``- decorated methods)

    :param bool unpack_extras:
        (default True) if True, any extra keyword arguments included
        in the decorator call will be passed as keyword arguments
        to ``func``. If ``False``, extra keyword arguments will be
        passed to func as a dict assigned to the keyword argument
        corresponding to ``key``

    :param str extras_key:
        (default ``"extras"``) the key to which to assign extra
        decorator keyword arguments when ``unpack_extras`` is ``False``

    :param bool _use_future_syntax:
        (default False) if True, use the new ``Decorated`` object,
        which will become the default in version 2.0.0. If this
        parameter is True, all other parameters will be ignored
        expect for ``instance_methods_only`` and
        ``implicit_method_decoration``, a ``Decorated`` object
        will be passed to the function as its first argument,
        and any provided keyword arguments will be unpacked.

    :param **dict extras:
        any extra keyword arguments supplied to the decoration call

    :return: the decorated function/method/class
    :rtype: Union[FunctionType,MethodType,type]
    """

    def decorator(decorated):
        """The function that returns a replacement for the original"""

        def wrapper(*args, **kwargs):
            """The function that replaces the decorated one"""

            fn = func

            if _use_future_syntax:
                decor = Decorated(decorated, args, kwargs)
                fn = partial(fn, decor)
            else:
                if pass_params:
                    fn_args = get_fn_args(decorated, args)
                    fn = partial(fn, fn_args, kwargs)

                if pass_decorated:
                    fn = partial(fn, decorated)

            if unpack_extras or _use_future_syntax:
                fkwargs = extras
            else:
                fkwargs = {extras_key: extras}

            return fn(**fkwargs)

        if implicit_method_decoration and isclass(decorated):
            return ClassWrapper.wrap(
                decorated,
                decorator,
                instance_methods_only=instance_methods_only
            )

        # Equivalent to @wraps(decorated) on `wrapper`
        return wraps(decorated)(wrapper)

    return decorator


def decorate(before=None, after=None, instead=None, before_opts=None,
             after_opts=None, instead_opts=None,
             implicit_method_decoration=True,
             instance_methods_only=False, unpack_extras=True,
             _use_future_syntax=False, extras_key='extras', **extras):
    """A decorator that combines before, after, and instead decoration

    The ``before``, ``after``, and ``instead`` decorators are all
    stackable, but this decorator provides a straightforward interface
    for combining them so that you don't have to worry about
    decorator precedence. In addition, this decorator ensures that
    ``instead`` is called *first*, so that it does not replace
    any desired calls to ``before`` or ``after``.

    This decorator takes three optional keyword arguments, ``before``,
    ``after``, and ``instead``, each of which may be any callable that
    might be provided to those decorators.

    The provided callables will be invoked with the same default
    call signature as for the individual decorators. Call signatures
    and other options may be adjusted by passing ``before_opts``,
    ``after_opts``, and ``instead_opts`` dicts to this decorator,
    which will be directly unpacked into the invocation of the
    individual decorators.

    The ``instance_methods_only`` and ``implicit_method_decoration``
    options may only be changed globally as arguments to this
    decorator, in order to avoid confusion between which decorator
    is being applied to which methods.

    You may specify decorator-specific extras in the various ``opts``
    dictionaries. In addition, any extra keyword arguments passed
    to this decorator will be passed to each of the provided
    callables. As with the individual decorators, if
    ``unpack_extras=False`` is provided, the extras will be packed
    into a dict and passed via the ``extras_key`` keyword argument.

    :param Callable before:
        a callable to run before the decorated function. By default,
        it is called with no arguments, but the call signature may
        be changed depending on the value of ``pass_params`` and
        ``pass_decorated``

    :param Callable after:
        a callable to run after the decorated function. By default,
        it is called with the result of the decorated function as
        its only argument, but the call signature may be changed
        depending on the value of ``pass_params`` and ``pass_decorated``

    :param Callable instead:
        a callable to run in place of the decorated function. By
        default, it is called the decorated function's arguments
        as a tuple, keyword arguments as a dict, and a reference
        to the decorated function

    :param dict before_opts:
        a dictionary of keyword arguments to pass to the ``before``
        decorator. See :any:`before` for supported options.

    :param dict after_opts:
        a dictionary of keyword arguments to pass to the ``after``
        decorator. See :any:`after` for supported options.

    :param dict instead_opts:
        a dictionary of keyword arguments to pass to the ``instead``
        decorator. See :any:`instead` for supported options

    :param bool implicit_method_decoration:
        (default True) if True, decorating a class implies decorating
        all of its methods. This value overrides any values set in
        the various ``opts`` dictionaries.

    :param bool instance_methods_only:
        (default False) if True, decorating a class implies decorating
        only its instance methods (not ``classmethod``- or
        ``staticmethod``- decorated methods). This value overrides
        any values set in the various ``opts`` dictionaries.

    :param bool unpack_extras:
        (default True) if True, any extra keyword arguments included
        in the decorator call will be passed as keyword arguments
        to ``func``. If ``False``, extra keyword arguments will be
        passed to func as a dict assigned to the keyword argument
        corresponding to ``key``. If provided, this value serves
        as the default for the various decorators.

    :param str extras_key:
        (default ``"extras"``) the key to which to assign extra
        decorator keyword arguments when ``unpack_extras`` is
        ``False``. If provided, this key serves as the default
        for the various decorators.

    :param bool _use_future_syntax:
        (default False) if True, use the new ``Decorated`` object,
        which will become the default in version 2.0.0. If this
        parameter is True, all other parameters will be ignored
        expect for ``instance_methods_only`` and
        ``implicit_method_decoration``, a ``Decorated`` object
        will be passed to the function as its first argument,
        and any provided keyword arguments will be unpacked.

    :param **dict extras:
        any extra keyword arguments supplied to the decoration call

    :return: the decorated function/method/class
    :rtype: Union[FunctionType,MethodType,type]
    """

    if all(arg is None for arg in (before, after, instead)):
        raise ValueError(
            'At least one of "before," "after," or "instead" must be provided'
        )

    my_before = before
    my_after = after
    my_instead = instead

    before_opts = before_opts or {}
    after_opts = after_opts or {}
    instead_opts = instead_opts or {}

    for opts in (before_opts, after_opts, instead_opts):
        # Disallow mixing of class-level functionality
        opts['implicit_method_decoration'] = implicit_method_decoration
        opts['instance_methods_only'] = instance_methods_only
        opts['_use_future_syntax'] = _use_future_syntax

    def decorator(decorated):

        wrapped = decorated

        if my_instead is not None:

            global instead
            dec_kwargs = _kwargs_from_opts(
                instead_opts, unpack_extras, extras_key, extras
            )
            wrapped = instead(my_instead, **dec_kwargs)(wrapped)

        if my_before is not None:

            global before
            dec_kwargs = _kwargs_from_opts(
                before_opts, unpack_extras, extras_key, extras
            )
            wrapped = before(my_before, **dec_kwargs)(wrapped)

        if my_after is not None:

            global after

            dec_kwargs = _kwargs_from_opts(
                after_opts, unpack_extras, extras_key, extras
            )
            wrapped = after(my_after, **dec_kwargs)(wrapped)

        def wrapper(*args, **kwargs):

            return wrapped(*args, **kwargs)

        if implicit_method_decoration and isclass(wrapped):
            return ClassWrapper.wrap(
                decorated,
                decorator,
                instance_methods_only=instance_methods_only
            )

        return wraps(decorated)(wrapper)

    return decorator


def construct_decorator(before=None, after=None, instead=None,
                        before_opts=None, after_opts=None, instead_opts=None,
                        implicit_method_decoration=True,
                        instance_methods_only=False, unpack_extras=True,
                        _use_future_syntax=False,
                        extras_key='extras', **extras):
    """Return a custom decorator

    Options are all the same as for :any:`decorate` and are, in fact,
    passed directly to it.

    Once the decorator has been created, it can be used like any
    other decorators in this module. Passing extra keyword arguments
    during the decoration call will pass those extras to the
    provided callables, even if some default extras were already set
    when creating the decorator.

    :return: a decorator that can be used to decorate functions,
        classes, or methods
    :rtype: FunctionType
    """
    return partial(
        decorate, before=before, after=after, instead=instead,
        before_opts=before_opts, after_opts=after_opts,
        instead_opts=instead_opts,
        implicit_method_decoration=implicit_method_decoration,
        instance_methods_only=instance_methods_only,
        unpack_extras=unpack_extras, extras_key=extras_key,
        _use_future_syntax=_use_future_syntax,
        **extras
    )


def _kwargs_from_opts(opts, unpack_extras, extras_key, extras):
    kwargs = opts
    if (opts.get('unpack_extras', unpack_extras) or
            opts.get('_use_future_syntax', False)):
        kwargs.update(extras)
    else:
        key = opts.get('extras_key', extras_key)
        kwargs[key] = extras
    return kwargs



# -*- coding: UTF-8 -*-
"""
Tests for caches
"""

from time import sleep

import pytest

from pydecor.caches import FIFOCache, LRUCache, TimedCache


class TestLRU:
    """Test the LRU cache"""

    def test_unsized(self):
        """Test that not specifying a size lets the cache grow"""
        cache = LRUCache()
        for i in range(500):
            cache[i] = i
        for i in range(500):
            assert i in cache
            assert cache[i] == i

    def test_sized_no_reuse(self):
        """Test basic LIFO functionality with all new values"""
        cache = LRUCache(max_size=5)
        for i in range(5):
            cache[i] = i
        for i in range(5):
            assert i in cache
            assert cache[i] == i
        for i in range(5, 10):
            cache[i] = i
            assert i in cache
            assert cache[i] == i
            assert i - 5 not in cache
            with pytest.raises(KeyError):
                assert cache[i - 5]

    def test_sized_with_reuse(self):
        """Test LRU functionality"""
        cache = LRUCache(max_size=3)
        for i in range(3):
            cache[i] = i
        # LRU: 0 1 2
        for i in range(3):
            assert i in cache
            assert cache[i] == i
        # LRU: 0 1 2

        cache[3] = 3
        # LRU: 1 2 3

        assert 0 not in cache
        with pytest.raises(KeyError):
            assert cache[0]

        for i in range(1, 4):
            assert i in cache
            assert cache[i] == i

        # Test re-ording with getitem
        assert cache[1]
        # LRU: 2 3 1

        cache[0] = 0
        # LRU: 3 1 0

        assert 2 not in cache
        with pytest.raises(KeyError):
            assert cache[2]

        for i in 3, 1, 0:
            assert i in cache
            assert cache[i] == i

        # Test re-ording with setitem
        cache[3] = 3
        # LRU: 1 0 3

        for i in 1, 0, 3:
            assert i in cache
            assert cache[i] == i

        cache[2] = 2
        # LRU: 0 3 2

        for i in 0, 3, 2:
            assert i in cache
            assert cache[i] == i

        assert 1 not in cache
        with pytest.raises(KeyError):
            assert cache[1]


class TestFIFO:
    """Test the LRU cache"""

    def test_unsized(self):
        """Test that not specifying a size lets the cache grow"""
        cache = FIFOCache()
        for i in range(500):
            cache[i] = i
        for i in range(500):
            assert i in cache
            assert cache[i] == i

    def test_sized_no_reuse(self):
        """Test basic LIFO functionality with all new values"""
        cache = FIFOCache(max_size=5)
        for i in range(5):
            cache[i] = i
        for i in range(5):
            assert i in cache
            assert cache[i] == i
        for i in range(5, 10):
            cache[i] = i
            assert i in cache
            assert cache[i] == i
            assert i - 5 not in cache
            with pytest.raises(KeyError):
                assert cache[i - 5]

    def test_sized_with_reuse(self):
        """Test LRU functionality"""
        cache = FIFOCache(max_size=3)
        for i in range(3):
            cache[i] = i
        # Cache: 0 1 2
        for i in range(3):
            assert i in cache
            assert cache[i] == i
        # Cache: 0 1 2

        cache[3] = 3
        # Cache: 1 2 3

        assert 0 not in cache
        with pytest.raises(KeyError):
            assert cache[0]

        for i in range(1, 4):
            assert i in cache
            assert cache[i] == i

        assert cache[1]
        # Cache: 1 2 3

        cache[0] = 0
        # Cache: 2 3 0

        assert 1 not in cache
        with pytest.raises(KeyError):
            assert cache[1]

        for i in 2, 3, 0:
            assert i in cache
            assert cache[i] == i


class TestTimedCache:

    def test_untimed(self):
        """Test that not specifying a size lets the cache grow"""
        cache = TimedCache()
        for i in range(500):
            cache[i] = i
        for i in range(500):
            assert i in cache
            assert cache[i] == i

    def test_timed(self):
        """Test that entries are removed if accessed after max_age"""
        time = 0.001
        cache = TimedCache(max_age=time)

        cache[1] = 1
        assert 1 in cache
        sleep(time)
        assert 1 not in cache
        with pytest.raises(KeyError):
            assert cache[1]

        for i in range(50):
            cache[i] = i
            assert i in cache
            assert cache[i] == i
        sleep(time)
        for i in range(50):
            assert i not in cache
            with pytest.raises(KeyError):
                assert cache[i]

    def test_timed_reset(self):
        """Test that resetting entries resets their time"""
        time = 0.005
        cache = TimedCache(max_age=time)

        cache[1] = 1
        assert 1 in cache
        assert cache[1] == 1
        sleep(time / 2)
        assert 1 in cache
        assert cache[1] == 1
        cache[1] = 1
        sleep(time / 2)
        assert 1 in cache
        assert cache[1] == 1
        sleep(time / 2)
        assert 1 not in cache
        with pytest.raises(KeyError):
            assert cache[1]



# -*- coding: UTF-8 -*-
"""
Functions to use for decorator construction
"""

from __future__ import absolute_import, unicode_literals


__all__ = (
    'intercept',
    'log_call',
)


from inspect import getmodule
from logging import getLogger
from six import raise_from
from sys import version_info


from .constants import LOG_CALL_FMT_STR
from ._memoization import convert_to_hashable


PY2 = version_info < (3, 0)


def intercept(decorated, catch=Exception, reraise=None, handler=None,
              err_msg=None, include_context=False):
    """Intercept an error and either re-raise, handle, or both

    Designed to be called via the ``instead`` decorator.

    :param Decorated decorated: decorated function information
    :param Type[Exception] catch: an exception to intercept
    :param Union[bool, Type[Exception]] reraise: if provided, will re-raise
        the provided exception, after running any provided
        handler callable. If ``False`` or ``None``, no exception
        will be re-raised.
    :param Callable[[Type[Exception]],Any] handler: a function
        to call with the caught exception as its only argument.
        If not provided, nothing will be done with the caught
        exception.
    :param str err_msg: if included will be used to instantiate
        the exception. If not included, the caught exception will
        be cast to a string and used instead
    :param include_context: if True, the previous exception will
        be included in the exception context.
    """
    try:
        return decorated(*decorated.args, **decorated.kwargs)

    except catch as exc:
        if handler is not None:
            handler(exc)

        if isinstance(reraise, bool):
            if reraise:
                raise

        elif reraise is not None:

            if err_msg is not None:
                new_exc = reraise(err_msg)
            else:
                new_exc = reraise(str(exc))

            context = exc if include_context else None
            raise_from(new_exc, context)


def log_call(decorated, logger=None, level='info',
             format_str=LOG_CALL_FMT_STR):
    """Log the parameters & results of a function call

    Designed to be called via the ``after`` decorator with
    ``pass_params=True`` and ``pass_decorated=True``. Use
    :any:`decorators.log_call` for easiest invocation.

    :param Decorated decorated: decorated function information
    :param Optional[logging.Logger] logger: optional logger instance
    :param Optional[str] level: log level - must be an acceptable Python
        log level, case-insensitive
    :param format_str: the string to use when formatting the results
    """
    if logger is None:
        name = getmodule(decorated.wrapped).__name__
        logger = getLogger(name)
    log_fn = getattr(logger, level.lower())
    msg = format_str.format(
        name=decorated.wrapped.__name__,
        args=decorated.args,
        kwargs=decorated.kwargs,
        result=decorated.result
    )
    log_fn(msg)


def memoize(decorated, memo):
    """Return a memoized result if possible; store if not present

    :param Decorator decorated: decorated function information
    :param memo: the memoization cache. Must support standard
        __getitem__ and __setitem__ calls
    """
    key = convert_to_hashable(decorated.args, decorated.kwargs)
    if key in memo:
        return memo[key]
    res = decorated(*decorated.args, **decorated.kwargs)
    memo[key] = res
    return res



# -*- coding: UTF-8 -*-
"""
Required functionality for the memoization function
"""

__all__ = (
    'convert_to_hashable',
    'hashable',
)

import dill as pickle


def convert_to_hashable(args, kwargs):
    """Return args and kwargs as a hashable tuple"""
    return hashable(args), hashable(kwargs)


def hashable(item):
    """Get return a hashable version of an item

    If the item is natively hashable, return the item itself. If
    it is not, return it dumped to a pickle string.
    """
    try:
        hash(item)
    except TypeError:
        item = pickle.dumps(item)
    return item



# -*- coding: UTF-8 -*-
"""
Tests for memoization functions
"""

import pytest

from pydecor._memoization import hashable


@pytest.mark.parametrize('item', [
    'foo',
    12,
    12.5,
    7e7,
    {'foo': 'bar'},
    object(),
    type('a', (object, ), {}),
    type('a', (object, ), {})(),
    lambda x: 'foo',
    {'a', 'b', 'c'},
    ('a', 'b', 'c'),
    ['a', 'b', 'c'],
    ('a', {'b': 'c'}, ['d']),
    ['a', ('b', 'c'), {'d': 'e'}],
])
def test_hashable(item):
    """Test getting a hashable verison of an item

    Asserts that the hash method does not error, which it does
    if the returned item is unhashable
    """
    hash(hashable(item))



# -*- coding: UTF-8 -*-
"""
setup module for ihiji_schemas
"""

from os.path import dirname, join, realpath
from setuptools import setup, find_packages
from textwrap import dedent


fdir = dirname(realpath(__file__))


NAME = 'pydecor'
URL = 'https://github.com/mplanchard/pydecor'
AUTHOR = 'Matthew Planchard'
EMAIL = 'msplanchard@gmail.com'


SHORT_DESC = 'Easy peasy Python decorators'

LONG_DESC = dedent(
    '''
    Decorators are great, but they're hard to write, especially if you want
    to include arguments to your decorators, or use your decorators on
    class methods as well as functions. I know that, no matter how many I 
    write, I still find myself looking up the syntax every time. And that's 
    just for simple function decorators. Getting decorators to work 
    consistently at the class and method level is a whole 'nother barrel 
    of worms.

    PyDecor aims to make function decoration easy and straightforward, so that
    developers can stop worrying about closures and syntax in triply nested
    functions and instead get down to decorating!
    '''
)


KEYWORDS = [
    'decorators',
    'python2',
    'python3',
]


PACKAGE_DEPENDENCIES = []
SETUP_DEPENDENCIES = []
TEST_DEPENDENCIES = []
EXTRAS_DEPENDENCIES = {}

ENTRY_POINTS = {}

with open(join(fdir, 'requirements.txt')) as reqfile:
    for ln in reqfile:
        if not ln.startswith('#'):
            PACKAGE_DEPENDENCIES.append(ln.strip())

# See https://pypi.python.org/pypi?%3Aaction=list_classifiers for all
# available setup classifiers
CLASSIFIERS = [
    # 'Development Status :: 1 - Planning',
    # 'Development Status :: 2 - Pre-Alpha',
    # 'Development Status :: 3 - Alpha',
    # 'Development Status :: 4 - Beta',
    'Development Status :: 5 - Production/Stable',
    # 'Development Status :: 6 - Mature',
    # 'Framework :: AsyncIO',
    # 'Framework :: Flask',
    # 'Framework :: Sphinx',
    # 'Environment :: Web Environment',
    'Intended Audience :: Developers',
    # 'Intended Audience :: End Users/Desktop',
    # 'Intended Audience :: Science/Research',
    # 'Intended Audience :: System Administrators',
    # 'License :: Other/Proprietary License',
    # 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    # 'Operating System :: MacOS :: MacOS X
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python',
    # 'Programming Language :: Python :: 3 :: Only',
    # 'Programming Language :: Python :: Implementation :: PyPy',
]


__version__ = '0.0.0'

with open(join(fdir, '{0}/_version.py'.format(NAME))) as version_file:
    for line in version_file:
        # This will populate the __version__ and __version_info__ variables
        if line.startswith('__'):
            exec(line)

setup(
    name=NAME,
    version=__version__,
    description=SHORT_DESC,
    long_description=LONG_DESC,
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=find_packages(exclude=['*.tests', '*.tests.*']),
    install_requires=PACKAGE_DEPENDENCIES,
    setup_requires=SETUP_DEPENDENCIES,
    tests_require=TEST_DEPENDENCIES,
    extras_require=EXTRAS_DEPENDENCIES,
)



# -*- coding: UTF-8 -*-
"""
Public interface for decorators sub-package
"""

__all__ = (
    'after',
    'before',
    'construct_decorator',
    'decorate',
    'instead',
    'Decorated',
    'DecoratorType',
    'intercept',
    'log_call',
    'memoize',
)

from .generic import *
from .ready_to_wear import *



# -*- coding: UTF-8 -*-
"""
Prête-à-porte (ready-to-wear) decorators are ready to be used immediately
in your codebase, minimal thinking required.
"""

from __future__ import absolute_import, unicode_literals


__all__ = (
    'intercept',
    'log_call',
    'memoize',
)


from logging import getLogger

from pydecor import functions
from pydecor.constants import LOG_CALL_FMT_STR
from pydecor.caches import LRUCache

from .generic import instead, after


log = getLogger(__name__)


def intercept(catch=Exception, reraise=None, handler=None, err_msg=None,
              include_context=False):
    """Intercept an exception and either re-raise, handle, or both

    Example:

    .. code:: python

        from logging import getLogger
        from pydecor import intercept

        log = getLogger(__name__)

        class MyLibraryError(Exception):
            pass

        def error_handler(exception):
            log.exception(exception)

        # Re-raise and handle
        @intercept(catch=ValueError, reraise=MyLibraryError,
                   handler=error_handler):
        def some_error_prone_function():
            return int('foo')

        # Re-raise only
        @intercept(catch=TypeError, reraise=MyLibraryError)
        def another_error_prone_function(splittable_string):
            return splittable_string.split('/', '.')

        # Handle only
        @intercept(catch=ValueError, handle=error_handler)
        def ignorable_error(some_string):
            # Just run the handler on error, rather than re-raising
            log.info('This string is {}'.format(int(string)))


    :param Type[Exception] catch: the exception to catch
    :param Union[bool, Type[Exception]] reraise: the exception to
        re-raise or ``True``. If an exception is provided, that
        exception will be raised. If ``True``, the original
        exception will be re-raised. If ``False`` or ``None``,
        no exception will be raised. Note that if a ``handler``
        is provided, it will always be called prior to re-raising.
    :param Callable[[Type[Exception]],Any] handler: a function to call
        with the caught exception as its only argument. If not provided,
        nothing will be done with the caught exception
    :param str err_msg: An optional string with which to call the
        re-raised exception. If not provided, the caught exception
        will be cast to a string and used instead.
    :param include_context: if True, the previous exception will be
        included in the context of the re-raised exception, which
        means the traceback will show both exceptions, noting that
        the first exception "was the direct cause of the following
        exception"
    """
    return instead(
        functions.intercept,
        _use_future_syntax=True,
        catch=catch,
        reraise=reraise,
        handler=handler,
        err_msg=err_msg,
        include_context=include_context,
    )


def log_call(logger=None, level='info', format_str=LOG_CALL_FMT_STR):
    """Log the name, parameters, & result of a function call

    If not provided, a logger instance will be retrieved corresponding
    to the module name of the decorated function, so if you decorate
    the function ``do_magic()`` in the module ``magic.py``, the
    retrieved logger will be the same as one retrieved in ``magic.py``
    with ``logging.getLogger(__name__)``

    The ``level`` provided here **does not** set the log level of the
    passed or retrieved logger. It just determines what level the
    generated message should be logged with.

    Example:

    * Assume the following is found in ``log_example.py``

    .. code:: python

        from pydecor import log_call

        @log_call
        def return_none(*args, **kwargs):
            return None

        return_none('alright', my='man')

        # Will retrieve the ``log_example`` logger
        # and log the message:
        # "return_none(*('alright', ), **{'my': 'man'}) -> None"

    :param Optional[logging.Logger] logger: an optional Logger
        instance. If not provided, the logger corresponding to the
        decorated function's module name will be retrieved
    :param str level: the level with which to log the message. Must
        be an acceptable Python log level
    :param str format_str: the format string to use when interpolating
        the message. This defaults to
        ``'{name}(*{args}, **{kwargs}) -> {result}'``. Any provided
        format string should contain the same keys, which will be
        interpolated appropriately.

    :rtype: DecoratorType
    """
    return after(
        functions.log_call,
        _use_future_syntax=True,
        logger=logger,
        level=level,
        format_str=format_str,
    )


def memoize(keep=0, cache_class=LRUCache):
    """Memoize the decorated function

    The default cache is an infinitely growing LRU Cache. To specify
    a maximum number of key/value pairs to keep, specify ``keep``.
    To specify a different cache, pass it to ``cache_class``. Any
    class that implements ``__contains__``, ``__getitem__``, and
    ``__setitem__`` may be used. as the cache. There are several
    cache types provided in `pydecor.caches`_.

    :param int keep: the maximum size of the cache (or max age of the
        cache entries in seconds when using TimedCache). By default
        this is 0, which means the cache can grow indefinitely.
    :param cache_class: the cache to store function results.
        Any class that supports __getitem__, __setitem__,
        and __contains__ should work just fine here. The
        max_size is passed in as an instantiation parameter,
        so that should also be supported. In addition to
        the default LRUCache, a LIFOCache and a TimedCache
        are provided in ``pydecor.caches``.

    :rtype: DecoratorType
    """
    return instead(
        functions.memoize,
        _use_future_syntax=True,
        memo=cache_class(keep),
    )
